#include "UG3_LIB.h"

/*
 * UG3 LIB : Unstructured Grid - General Purpose Routine Library
 * 3D Version : $Id: ug3_conv_2d_mesh_3d.c,v 1.7 2023/11/01 04:36:14 marcum Exp $
 * Copyright 1994-2021, David L. Marcum
 */

INT_ ug3_conv_2d_mesh_3d (
  INT_ mesh_3d,
  INT_ mmsg,
  INT_ nplane,
  double dang,
  double dz,
  INT_ nbedge_2d,
  INT_ ntria_2d,
  INT_ nquad_2d,
  INT_ nnode_2d,
  INT_1D *idibe_2d,
  INT_2D *inibe_2d,
  INT_4D *iniq_2d,
  INT_3D *init_2d,
  DOUBLE_2D *x_2d,
  INT_ *nbface,
  INT_ *nelem,
  INT_ *nnode,
  INT_1D **idibf,
  INT_3D **inibf,
  INT_4D **iniel,
  DOUBLE_3D **x)
{
  // Convert a 2D planar mesh to an axisymmetric or stacked plane 3D mesh.

  // mesh_3d=0	: do nothing
  // mesh_3d=1	: convert to a axisymmetric 3D volume mesh
  // mesh_3d=2	: convert to a stacked plane 3D volume mesh

  // dang	: angle between rotated planes with axisymmetric volume mesh
  // dz		: distance between planes with stacked plane volume mesh

  // nplanes	: number of planes for 3D volume mesh

  INT_1D *itin_2d = NULL;
  INT_1D *jnin = NULL;
  INT_1D *litin_2d = NULL;
  INT_1D *plane_in_2d = NULL;

  DOUBLE_3D *x_save = NULL;

  INT_ i, i1, i2, i3, i4, ibedge_2d, ibface, ielem, inode, inode_2d,
       jbface, jelem, jnode, loc, ncl, plane;
  INT_ id_max = 1;
  INT_ ierr = 0;
  INT_ nbfacei = 0;
  INT_ nbfpnt = 0;
  INT_ nbfpntd = 0;
  INT_ nnodeb_2d = 0;
  INT_ nquad = 0;

  double ang, angi, dang_;

  *nelem = 0;
  *nnode = 0;

  *idibf = NULL;
  *inibf = NULL;
  *iniel = NULL;
  *x = NULL;

  // exit if 3D volume mesh option is off

  if (mesh_3d == 0) return 0;

  // set number of boundary edge nodes

  ug2_nnodeb (nbedge_2d, &nnodeb_2d, inibe_2d);

  // allocate boundary face connectivity for 3D volume mesh 

  *nbface = ntria_2d;

  *inibf = (INT_3D *) ug_malloc (&ierr, ((*nbface)+1) * sizeof (INT_3D));

  if (ierr) {
    ug_error_message ("*** ERROR 100207 : unable to allocate required memory ***");
    ierr = 100207;
    goto cleanup;
  }

  // initialize boundary face connectivity for 3D volume mesh 

  for (ibface = 1; ibface <= ntria_2d; ibface++) {
    (*inibf)[ibface][0] = init_2d[ibface][0];
    (*inibf)[ibface][1] = init_2d[ibface][1];
    (*inibf)[ibface][2] = init_2d[ibface][2];
  }

  // convert quad faces to tria faces

  nquad = nquad_2d;

  ierr = ug2_qtria (nbface, &nquad, inibf, iniq_2d, x_2d);

  if (ierr) goto cleanup;

  // allocate output and work data

  nbfacei = *nbface;

  *nbface = 2*nbfacei + 2*(nplane-1)*nnodeb_2d;
  *nelem = 3*(nplane-1)*nbfacei;
  *nnode = nplane * nnode_2d;

  dang_ = dang * atan(1.0) / 45.0;
  angi = atan(1.0) + atan(1.0);

  *idibf = (INT_1D *) ug_malloc (&ierr, ((*nbface)+1) * sizeof (INT_1D));
  *inibf = (INT_3D *) ug_realloc (&ierr, *inibf, ((*nbface)+1) * sizeof (INT_3D));
  *iniel = (INT_4D *) ug_malloc (&ierr, ((*nelem)+1) * sizeof (INT_4D));

  litin_2d = (INT_1D *) ug_malloc (&ierr, (nnode_2d+2) * sizeof (INT_1D));
  plane_in_2d = (INT_1D *) ug_malloc (&ierr, (nnode_2d+1) * sizeof (INT_1D));

  *x = (DOUBLE_3D *) ug_malloc (&ierr, ((*nnode)+1) * sizeof (DOUBLE_3D));

  if (ierr) {
    ug_error_message ("*** ERROR 100207 : unable to allocate required memory ***");
    ierr = 100207;
    goto cleanup;
  }

  // allocate work data

  memset (plane_in_2d, 0, (nnode_2d+1)*sizeof (INT_1D));

  // set map of boundary faces surrounding each node

  ierr = ug2_ielin (1, nbfacei, &nbfpntd, nnode_2d, &nbfpnt, 
                    *inibf, NULL, &itin_2d, litin_2d);

  if (ierr) goto cleanup;

  // set coordinates for axisymmetric 3D volume mesh

  if (mesh_3d == 1) {

    for (plane = 0; plane < nplane; plane++) {

      ang = angi + dang_ * (double) plane;

      for (inode_2d = 1; inode_2d <= nnode_2d; inode_2d++) {

        inode = inode_2d + plane * nnode_2d;

        (*x)[inode][0] = x_2d[inode_2d][0];
        (*x)[inode][1] = x_2d[inode_2d][1] * sin(ang);
        (*x)[inode][2] = x_2d[inode_2d][1] * cos(ang);
      }
    }
  }

  // set coordinates for stacked plane 3D volume mesh

  else {

    for (plane = 0; plane < nplane; plane++) {

      for (inode_2d = 1; inode_2d <= nnode_2d; inode_2d++) {

        inode = inode_2d + plane * nnode_2d;

        (*x)[inode][0] = x_2d[inode_2d][0];
        (*x)[inode][1] = x_2d[inode_2d][1];
        (*x)[inode][2] = - dz * plane;
      }
    }
  }

  // set boundary face connectivity and ID for 3D volume mesh

  if (idibe_2d) id_max = ug_max_int (1, nbedge_2d, idibe_2d);

  for (ibface = 1; ibface <= nbfacei; ibface++) {

    (*idibf)[ibface] = id_max+1;
    (*idibf)[ibface+nbfacei] = id_max+2;

    i2 = (*inibf)[ibface][1];
    i3 = (*inibf)[ibface][2];

    (*inibf)[ibface+nbfacei][0] = (*inibf)[ibface][0] + (nplane-1) * nnode_2d;
    (*inibf)[ibface+nbfacei][1] = (*inibf)[ibface][1] + (nplane-1) * nnode_2d;
    (*inibf)[ibface+nbfacei][2] = (*inibf)[ibface][2] + (nplane-1) * nnode_2d;
    (*inibf)[ibface][1] = i3;
    (*inibf)[ibface][2] = i2;
  }

  ibface = 2*nbfacei;

  for (plane = 1; plane < nplane; plane++) {

    for (ibedge_2d = 1; ibedge_2d <= nbedge_2d; ibedge_2d++) {

      i1 = inibe_2d[ibedge_2d][0];
      i2 = inibe_2d[ibedge_2d][1];

      if (i1 < i2) {

        ibface++;

        (*idibf)[ibface] = (idibe_2d) ? idibe_2d[ibedge_2d]: 1;
        (*inibf)[ibface][0] = i1 + (plane-1) * nnode_2d;
        (*inibf)[ibface][1] = i2 + (plane-1) * nnode_2d;
        (*inibf)[ibface][2] = i1 + plane * nnode_2d;

        ibface++;

        (*idibf)[ibface] = (idibe_2d) ? idibe_2d[ibedge_2d]: 1;
        (*inibf)[ibface][0] = i2 + (plane-1) * nnode_2d;
        (*inibf)[ibface][1] = i2 + plane * nnode_2d;
        (*inibf)[ibface][2] = i1 + plane * nnode_2d;
      }

      else {

        ibface++;

        (*idibf)[ibface] = (idibe_2d) ? idibe_2d[ibedge_2d]: 1;
        (*inibf)[ibface][0] = i1 + (plane-1) * nnode_2d;
        (*inibf)[ibface][1] = i2 + (plane-1) * nnode_2d;
        (*inibf)[ibface][2] = i2 + plane * nnode_2d;

        ibface++;

        (*idibf)[ibface] = (idibe_2d) ? idibe_2d[ibedge_2d]: 1;
        (*inibf)[ibface][0] = i1 + (plane-1) * nnode_2d;
        (*inibf)[ibface][1] = i2 + plane * nnode_2d;
        (*inibf)[ibface][2] = i1 + plane * nnode_2d;
      }
    }
  }

  // set element connectivity and ID for 3D volume mesh

  ielem = 0;

  for (plane = 0; plane < nplane-1; plane++) {

    for (inode_2d = 1; inode_2d <= nnode_2d; inode_2d++) {

      for (loc = litin_2d[inode_2d]; loc < litin_2d[inode_2d+1]; loc++) {

        ibface = itin_2d[loc];

        if (inode_2d == (*inibf)[ibface][0]) {
          i1 = inode_2d;
          i2 = (*inibf)[ibface][1];
          i3 = (*inibf)[ibface][2];
        }
        else if (inode_2d == (*inibf)[ibface][1]) {
          i1 = inode_2d;
          i2 = (*inibf)[ibface][2];
          i3 = (*inibf)[ibface][0];
        }
        else {
          i1 = inode_2d;
          i2 = (*inibf)[ibface][0];
          i3 = (*inibf)[ibface][1];
        }

        ielem++;

        (*iniel)[ielem][0] = i1 + plane * nnode_2d;
        (*iniel)[ielem][1] = (i1 > i2) ? i2 + plane * nnode_2d:
                                         i2 + (plane+1) * nnode_2d;
        (*iniel)[ielem][2] = (i1 > i3) ? i3 + plane * nnode_2d:
                                         i3 + (plane+1) * nnode_2d;
        (*iniel)[ielem][3] = i1 + (plane+1) * nnode_2d;

        (plane_in_2d[inode_2d])++;
      }
    }
  }

  // remove duplicate coordinates in axisymmetric 3D volume mesh

  if (mesh_3d == 1) {

    // allocate work data

    jnin = (INT_1D *) ug_malloc (&ierr, ((*nnode)+1) * sizeof (INT_1D));
    x_save = (DOUBLE_3D *) ug_malloc (&ierr, ((*nnode)+1) * sizeof (DOUBLE_3D));

    if (ierr) {
      ug_error_message ("*** ERROR 100207 : unable to allocate required memory ***");
      ierr = 100207;
      goto cleanup;
    }

    // flag all nodes on center line

    memset (jnin, 0, ((*nnode)+1)*sizeof (INT_1D));

    ncl = 0;

    for (inode = 1; inode <= nnode_2d; inode++) {

      if ((*x)[inode][1] == 0.0) {

        ncl++;

        jnin[inode] = 1;
      }
    }

    // remove duplicate nodes on center line

    jnode = 0;

    for (inode = 1; inode <= nnode_2d; inode++) {

      if (jnin[inode] == 1) {

        jnode++;

        jnin[inode] = jnode;

        x_save[jnode][0] = (*x)[inode][0];
        x_save[jnode][1] = (*x)[inode][1];
        x_save[jnode][2] = (*x)[inode][2];

        for (plane = 1; plane < nplane; plane++)
          jnin[inode+plane*nnode_2d] = jnode;
      }
    }

    for (inode = 1; inode <= *nnode; inode++) {

      if (jnin[inode] == 0) {

        jnode++;

        jnin[inode] = jnode;

        x_save[jnode][0] = (*x)[inode][0];
        x_save[jnode][1] = (*x)[inode][1];
        x_save[jnode][2] = (*x)[inode][2];
      }
    }

    *nnode = jnode;

    for (inode = 1; inode <= *nnode; inode++) {
      (*x)[inode][0] = x_save[inode][0];
      (*x)[inode][1] = x_save[inode][1];
      (*x)[inode][2] = x_save[inode][2];
    }

    // reset boundary face connectivity

    jbface = 0;

    for (ibface = 1; ibface <= *nbface; ibface++) {

      i = 0;

      if (jnin[(*inibf)[ibface][0]] <= ncl) i++;
      if (jnin[(*inibf)[ibface][1]] <= ncl) i++;
      if (jnin[(*inibf)[ibface][2]] <= ncl) i++;

      if (i < 3) {

        jbface++;

        (*idibf)[jbface] = (*idibf)[ibface];

        (*inibf)[jbface][0] = jnin[(*inibf)[ibface][0]];
        (*inibf)[jbface][1] = jnin[(*inibf)[ibface][1]];
        (*inibf)[jbface][2] = jnin[(*inibf)[ibface][2]];
      }
    }

    *nbface = jbface;

    // reset element connectivity

    jelem = 0;

    for (ielem = 1; ielem <= *nelem; ielem++) {

      i1 = jnin[(*iniel)[ielem][0]];
      i2 = jnin[(*iniel)[ielem][1]];
      i3 = jnin[(*iniel)[ielem][2]];
      i4 = jnin[(*iniel)[ielem][3]];

      if (i1 != i2 && i1 != i3 && i1 != i4 && 
          i2 != i3 && i2 != i4 && i3 != i4) {

        i = 0;

        if (i1 <= ncl) i++;
        if (i2 <= ncl) i++;
        if (i3 <= ncl) i++;
        if (i4 <= ncl) i++;

        if (i < 3) {

          jelem++;

          (*iniel)[jelem][0] = i1;
          (*iniel)[jelem][1] = i2;
          (*iniel)[jelem][2] = i3;
	  (*iniel)[jelem][3] = i4;
        }
      }
    }

    *nelem = jelem;
  }

  cleanup:

  // cleanup work data

  ug_free (itin_2d);
  ug_free (jnin);
  ug_free (litin_2d);
  ug_free (plane_in_2d);
  ug_free (x_save);

  return (ierr);
}
